"""Translated from original Chinese version / 译自中文原文:

python项目有两个极端：
一种是诗，一种是屎。
写诗的人可能看不起写屎的，
但是实际上也只是样貌不同罢了。
诗虽然整洁、美丽，但是读起来很费劲，没有注释那也是赤石。
屎就更离谱了，全是堆叠出来的，结构复杂的，能跑的，
但是我没说是谁能跑，总之写的时候是爽的。
两者都是写完过一会儿后没人能读懂的，或者是很费劲才能读懂的，
能直接读懂的我觉得已经飞升了。
如果你知道自己在做什么，
那就自己管理吧。

能跑就行了，我说我

—— Original text by LittleWhiteCloud / 中文原文作者：小白云"""

THE_POEM = """
There are two extremes in Python, 
one is art while the other is fart. 
Those who create art look down on those who fart, 
while they are not actually far apart. 

Art is filled with beauty making it murky. 
Art is just fart with comments and space bars. 
Fart is whatever piled up from nothing, 
messy but runs fast, 
except it is the programmer who runs away fast, 
while anyway (s)he types fast. 

Both art and fart aren't easy to parse; 
those who understand are a rare class. 

For those who want to make arts with arts... 
...or maybe fart: 
If you know what you are doing, 
then do it yourself. 

Never mind, 
it will be OK if it runs fast... 
I mean myself...

—— Translated by rgzz666 from original text by LittleWhiteCloud
"""


print(THE_POEM)


# The Charmy development team / Charmy 开发团队
# Yours truely / 敬上
#
# 2026/02/05 (Y/M/D)
